"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.VersionsDatasource = exports.memCache = void 0;
const error_messages_1 = require("../../../constants/error-messages");
const logger_1 = require("../../../logger");
const external_host_error_1 = require("../../../types/errors/external-host-error");
const date_1 = require("../../../util/date");
const http_1 = require("../../../util/http");
const regex_1 = require("../../../util/regex");
const datasource_1 = require("../datasource");
exports.memCache = new Map();
class VersionsDatasource extends datasource_1.Datasource {
    id;
    constructor(id) {
        super(id);
        this.id = id;
    }
    getRegistryCache(registryUrl) {
        const cacheKey = `rubygems-versions-cache:${registryUrl}`;
        const regCache = exports.memCache.get(cacheKey) ?? {
            lastSync: new Date('2000-01-01'),
            packageReleases: new Map(),
            contentLength: 0,
            isSupported: false,
            registryUrl,
        };
        exports.memCache.set(cacheKey, regCache);
        return regCache;
    }
    async getReleases({ registryUrl, packageName, }) {
        logger_1.logger.debug(`getRubygemsOrgDependency(${packageName})`);
        // istanbul ignore if
        if (!registryUrl) {
            return null;
        }
        const regCache = this.getRegistryCache(registryUrl);
        await this.syncVersions(regCache);
        if (!regCache.isSupported) {
            throw new Error(error_messages_1.PAGE_NOT_FOUND_ERROR);
        }
        const versions = regCache.packageReleases.get(packageName);
        if (!versions) {
            return null;
        }
        const releases = versions.map((version) => ({ version }));
        return { releases };
    }
    /**
     * https://bugs.chromium.org/p/v8/issues/detail?id=2869
     */
    static copystr(x) {
        return (' ' + x).slice(1);
    }
    async updateRubyGemsVersions(regCache) {
        const url = `${regCache.registryUrl}/versions`;
        const options = {
            headers: {
                'accept-encoding': 'gzip',
                range: `bytes=${regCache.contentLength}-`,
            },
        };
        let newLines;
        try {
            logger_1.logger.debug('Rubygems: Fetching rubygems.org versions');
            const startTime = Date.now();
            newLines = (await this.http.get(url, options)).body;
            const durationMs = Math.round(Date.now() - startTime);
            logger_1.logger.debug(`Rubygems: Fetched rubygems.org versions in ${durationMs}`);
            regCache.isSupported = true;
        }
        catch (err) /* istanbul ignore next */ {
            if (err instanceof http_1.HttpError && err.response?.statusCode === 404) {
                regCache.isSupported = false;
                return;
            }
            if (err.statusCode !== 416) {
                regCache.contentLength = 0;
                regCache.packageReleases = new Map();
                logger_1.logger.debug({ err }, 'Rubygems fetch error');
                throw new external_host_error_1.ExternalHostError(new Error('Rubygems fetch error'));
            }
            logger_1.logger.debug('Rubygems: No update');
            regCache.lastSync = new Date();
            return;
        }
        for (const line of newLines.split(regex_1.newlineRegex)) {
            this.processLine(regCache, line);
        }
        regCache.lastSync = new Date();
    }
    processLine(regCache, line) {
        let split;
        let pkg;
        let versions;
        try {
            const l = line.trim();
            if (!l.length || l.startsWith('created_at:') || l === '---') {
                return;
            }
            split = l.split(' ');
            [pkg, versions] = split;
            pkg = VersionsDatasource.copystr(pkg);
            let existingVersions = regCache.packageReleases.get(pkg) ?? [];
            const lineVersions = versions.split(',').map((version) => version.trim());
            for (const lineVersion of lineVersions) {
                if (lineVersion.startsWith('-')) {
                    const deletedVersion = lineVersion.slice(1);
                    logger_1.logger.trace({ pkg, deletedVersion }, 'Rubygems: Deleting version');
                    existingVersions = existingVersions.filter((version) => version !== deletedVersion);
                }
                else {
                    existingVersions.push(VersionsDatasource.copystr(lineVersion));
                }
            }
            regCache.packageReleases.set(pkg, existingVersions);
        }
        catch (err) /* istanbul ignore next */ {
            logger_1.logger.warn({ err, line, split, pkg, versions }, 'Rubygems line parsing error');
        }
    }
    isDataStale({ lastSync }) {
        return (0, date_1.getElapsedMinutes)(lastSync) >= 15;
    }
    updateRubyGemsVersionsPromise = null;
    async syncVersions(regCache) {
        if (this.isDataStale(regCache)) {
            this.updateRubyGemsVersionsPromise =
                this.updateRubyGemsVersionsPromise ??
                    this.updateRubyGemsVersions(regCache);
            await this.updateRubyGemsVersionsPromise;
            this.updateRubyGemsVersionsPromise = null;
        }
    }
}
exports.VersionsDatasource = VersionsDatasource;
//# sourceMappingURL=versions-datasource.js.map